/**
 * Generate a fantasy-world map
 * License:MIT
 * via https://github.com/amitp/mapgen
 * @author amitp(amitp@cs.stanford.edu), original creator.
 * @author dashrava(dashrava.com), modified by / converting from terrain map to planet gen.
 */

package;


import com.dashrava.utils.GraphicUtils;
import com.dashrava.utils.Random;
import openfl.display.BitmapData;
import openfl.display.Graphics;
import openfl.display.Sprite;
import openfl.display.GradientType;
import openfl.geom.Matrix;
import openfl.geom.Point;

class PlanetGenerator extends Sprite {

	public var light:Point=new Point();
	public var pixelScale:Int=4;
	public var planetname:String;

	public var SEED:Int=42;
	public var OCEAN_ALTITUDE:Int=1;
	public var SIZE:Int=64;

	public var map:TerrainMap;

	public var detailMap:Sprite;
	public var overlayMap:Sprite;
	public var starMap:Sprite;
	public var atmos:Sprite;

	public var COL_SEA:Int=0x000099;
	public var COL_SHORE:Int=0xae8c4c;
	public var COL_SNOW:Int=0xffffff;
	public var COL_WET:Int=0x56821b;
	public var COL_MOIST:Int=0x3b8c43;
	public var COL_NORMAL:Int=0x54653c;
	public var COL_PARCHED:Int=0x334021;
	public var COL_DRY:Int=0x989a2d;
	public var COL_DEAD:Int=0xc2bd8c;

	public var smoothEnabled:Bool=true;		// Tile smoothing method using corner adjusting magic
	public var planetmaskEnabled:Bool=true;	// Round the planet tilemap with a mask
	public var atmosphereEnabled:Bool=true;
	public var starfieldEnabled:Bool=true;
	public var cloudsEnabled:Bool=true;
	public var debrisEnabled:Bool=true;
	public var bandedEnabled:Bool=true;		// Use banded light instead of a tiled gradient?

	// See Main for suggested ranges. Some of this is super arbitrary.
	public var cloudDensity:Float=0.2;
	public var debrisDensity:Float=0.2;
	public var debrisGap:Float=0.2;
	public var debrisField:Float=0.2;
	public var lightRadius:Float=0.5;
	public var atmosIntensity:Float=0.3;
	public var atmosRadius:Float=1.1;
	public var lightIntensity:Float=1;
	public var lightMinimum:Float=0.5;

	public function new(){

		super();

		map = new TerrainMap(SIZE, SEED);

		roll();
	}

	public function roll():Void {

		Random.seed=SEED;

		planetname=generateName();

		SIZE = Std.int(Random.value()* 64 + 32);

		COL_SEA=randomColour();
		COL_SHORE=randomColour();
		COL_SNOW=randomColour();
		COL_WET=randomColour();
		COL_MOIST=randomColour();
		COL_NORMAL=randomColour();
		COL_PARCHED=randomColour();
		COL_DRY=randomColour();
		COL_DEAD=randomColour();

		map = new TerrainMap(SIZE, SEED);
		map.generate();

		if(detailMap 	!= null && contains(detailMap))		removeChild(detailMap);
		if(starMap 		!= null && contains(starMap))		removeChild(starMap);
		if(overlayMap 	!= null && contains(overlayMap))	removeChild(overlayMap);

		detailMap=new Sprite();
		overlayMap=new Sprite();
		starMap=new Sprite();

		//detailMap.cacheAsBitmap = true;
		//overlayMap.cacheAsBitmap = true;
		//starMap.cacheAsBitmap = true;

		// Starfield
		generateStars();

		if (atmos != null){
			removeChild(atmos);
		}

		generateDetailMap(detailMap.graphics, overlayMap.graphics);

		var atmoSize:Float=detailMap.width * 0.5 * atmosRadius;

		var mat:Matrix=new Matrix();
		mat.createGradientBox(SIZE * lightRadius * 2, SIZE * lightRadius * 2, 0, light.x, light.y);

		atmos=new Sprite();

		var ratios:Array<Int> = [Std.int(50 + Random.value() * Random.value() * 150), 255];
		var alphas:Array<Float> = [0.1 + atmosIntensity, atmosIntensity * 0.5 + 0.05];

		if (atmosphereEnabled) {
			atmos.graphics.beginGradientFill(GradientType.RADIAL, [COL_SEA, GraphicUtils.scalehex(COL_SEA, 0.6)], alphas, ratios, mat);
			atmos.graphics.drawCircle(SIZE * 0.5 * pixelScale, SIZE * 0.5 * pixelScale, atmoSize);
		}

		if(planetmaskEnabled){
			var pmask:Sprite=new Sprite();
			pmask.name="planetmask";
			pmask.graphics.lineStyle(1, 0xFFFFFF);
			pmask.graphics.beginFill(0xFF0000, 0.1);
			pmask.graphics.drawCircle(SIZE * 0.5 * pixelScale, SIZE * 0.5 * pixelScale, detailMap.width * 0.45);

			detailMap.addChild(pmask);
			detailMap.mask=pmask;
		}

		//detailMap.scaleX=detailMap.scaleY 	= pixelScale;
		//overlayMap.scaleX=overlayMap.scaleY = pixelScale;
		//atmos.scaleX=atmos.scaleY 			= pixelScale;

		starMap.visible=starfieldEnabled;

		addChild(starMap);
		addChild(detailMap);
		addChild(overlayMap);
		addChild(atmos);

		detailMap.x 	=detailMap.y 	=-SIZE*0.5*pixelScale;
		overlayMap.x 	=overlayMap.y 	=-SIZE*0.5*pixelScale;
		atmos.x 		=atmos.y 		=-SIZE*0.5*pixelScale;

	}

	private function generateStars():Void {

		var stars:Int = Std.int(Random.value()* 500 + 100);

		var scol:Int;

		var starx:Float;
		var stary:Float;
		var starr:Float;

		for(j in 0... stars){

			var rR:Int = Std.int(90 + Math.random()* 50);
			var rB:Int = Std.int(rR - Math.random()* 50);
			var rG:Int = rR;

			scol=GraphicUtils.RGBToHex(rR, rG, rB);

			starx=Random.value()* 1280 - 1280 * 0.5;
			stary=Random.value()* 720  -  720 * 0.5;
			starr=Math.random()* Math.random()* 4;

			starMap.graphics.beginFill(scol, Math.random());

			if(smoothEnabled){
				starMap.graphics.drawCircle(starx, stary, starr);
			}
			else {
				starMap.graphics.drawRect(starx, stary, starr, starr);
			}
		}
	}

	private function generateName():String {

		var chars:Array<Dynamic>="aeiouyaeiouyaeiouyabcdefghijklmnopqrstuvxyz".split("");

		var chipname:String="";

		// Make a random string
		var l:Int=Std.int(4 + Random.value()* Random.value()* 20);

		for(i in 0...l){
			chipname +=chars[Std.int(Random.value()* chars.length)];
		}

		// Add a hyphen
		if(Random.value()>0.8){
			var sub:Int = Std.int(Random.value()* chipname.length-2);
			sub +=2;
			chipname=chipname.substr(0, sub)+ "-" + chipname.substr(sub);
		}

		// Add a number
		if(Random.value()>0.8){
			chipname=chipname + " " + Std.string(Std.int(Random.value()* Random.value()* 999));
		}

		return chipname;
	}

	function randomColour():Int {
		var rR:Int = Std.int(60 + Random.value()* 120);
		var rG:Int = Std.int(60 + Random.value()* 120);
		var rB:Int = Std.int(60 + Random.value()* 120);

		return GraphicUtils.RGBToHex(rR, rG, rB);
	}

	public function moistureAndAltitudeToColor(m:Float, a:Float):Int {
		var color:Int=0xff0000;

		if(a<OCEAN_ALTITUDE){
			color=COL_SEA;
		} else if(a<OCEAN_ALTITUDE + 5){
			color=GraphicUtils.scalehex(COL_WET, 0.8);
		} else if(m>200)
			color=COL_WET;
		else if(m>150)
			color=GraphicUtils.scalehex(COL_WET, 0.9);
		else if(m>100)
			color=GraphicUtils.scalehex(COL_WET, 0.8);
		else if(m>50)
			color=GraphicUtils.scalehex(COL_NORMAL, 1.1);
		else if(m>20)
			color=GraphicUtils.scalehex(COL_NORMAL, 1.2);
		else
			color=GraphicUtils.scalehex(COL_NORMAL, 1.3);

		/*else if(a>200){
		   color=GraphicUtils.scalehex(COL_SNOW, a / 255);
		 }*/

		return color;
	}

	private function generateDetailMap(planet:Graphics, overlay:Graphics):Void {

		var centerX:Float=SIZE * 0.5;
		var centerY:Float=SIZE * 0.5;

		// Parameters -- These seem to mess up on openfl windows target, adding weird lines between the tiles
		var cornerAdjust:Float;
		var randomAdjust:Float= 0;

		if(smoothEnabled){
			cornerAdjust=0.25;
		}
		else {
			cornerAdjust=0;
		}

		// We are drawing an area DETAILSIZE x DETAILSIZE.
		var x:Int, y:Int;
		planet.clear();
		overlay.clear();

		// Coordinates of the detail area:
		var boundsBottom:Int = SIZE;
		var boundsTop	:Int = 0;
		var boundsLeft	:Int = 0;
		var boundsRight	:Int = SIZE;

		// Make sure that we're entirely within the bounds of the map
		//bounds=bounds.intersection(new Rectangle(0, 0, SIZE, SIZE));

		// 2d Array of vertices
		var vertices:Array<Array<Point>> = new Array<Array<Point>>();
		for(x in boundsLeft ... boundsRight){
			vertices[x] = new Array<Point>();
			for(y in boundsTop ... boundsBottom){
				// TODO:we'd save a lot of allocation if we reused these points
				vertices[x][y]=new Point(x, y);
			}
		}

		// Move vertices randomly
		if(randomAdjust>0){
			for(x in boundsLeft ... boundsRight){
				for(y in boundsTop ... boundsBottom){
					var rand1:Float=Random.value() - 0.5;
					var rand2:Float=Random.value() - 0.5;
					vertices[x][y].x +=rand1 * randomAdjust;
					vertices[x][y].y +=rand2 * randomAdjust;
				}
			}
		}

		if(cornerAdjust>0){
			// Alter vertices if 3 of 4 squares has same type
			for(x in boundsLeft + 1 ... boundsRight){
				for(y in boundsTop + 1 ... boundsBottom){

					// Sprites at the four squares touching this vertex
					var cTL:Int=moistureAndAltitudeToColor(map.moisture[x - 1][y - 1], map.altitude[x - 1][y - 1]);
					var cTR:Int=moistureAndAltitudeToColor(map.moisture[x][y - 1], map.altitude[x][y - 1]);
					var cBL:Int=moistureAndAltitudeToColor(map.moisture[x - 1][y], map.altitude[x - 1][y]);
					var cBR:Int=moistureAndAltitudeToColor(map.moisture[x][y], map.altitude[x][y]);

					// Figure out which corner is odd, if any
					if(cTR==cBL && cBL==cBR && cTL !=cTR){ // TL is odd
						vertices[x][y].x -=cornerAdjust;
						vertices[x][y].y -=cornerAdjust;
					}
					if(cTL==cTR && cTR==cBL && cBL !=cBR){ // BR is odd
						vertices[x][y].x +=cornerAdjust;
						vertices[x][y].y +=cornerAdjust;
					}
					if(cTL==cBL && cBL==cBR && cTL !=cTR){ // TR is odd
						vertices[x][y].x +=cornerAdjust;
						vertices[x][y].y -=cornerAdjust;
					}
					if(cTL==cTR && cTR==cBR && cTL !=cBL){ // BL is odd
						vertices[x][y].x -=cornerAdjust;
						vertices[x][y].y +=cornerAdjust;
					}
						// TODO:this seems like stupid repetitive code TODO:what
						// should we do when we have tiles A A B C(A and A are
						// adjacent but only two of them)?
				}
			}
		}

		// Draw grid
		var g:Graphics;

		var m:Matrix=new Matrix();

		for(x in boundsLeft ... boundsRight - 1){
			for(y in boundsTop ... boundsBottom - 1){

				var vx:Float=vertices[x][y].x;
				var vy:Float=vertices[x][y].y;

				var dx:Float=(vx - SIZE * 0.5)*(vx - SIZE * 0.5);
				var dy:Float=(vy - SIZE * 0.5)*(vy - SIZE * 0.5);

				if(dx + dy>SIZE * 0.50 * SIZE * 0.50){
					continue;// don't draw stuff in the corners
				}

				// Separating zone between planet and debris
				if(dx + dy>SIZE * debrisField * SIZE * debrisField){

					if(dx + dy<SIZE *(debrisField+debrisGap)* SIZE *(debrisField+debrisGap)){
						continue;
					}
				}

				// debris density
				if(dx + dy>SIZE * debrisField * SIZE * debrisField){

					if(!debrisEnabled)continue;

					if(Math.random()>debrisDensity){
						continue;
					}
				}

				// TODO:we should have cached the tile ids when we generate the map...
				var c:Int=moistureAndAltitudeToColor(map.moisture[x][y], map.altitude[x][y]);

				var scal:Float=((SIZE * lightRadius * SIZE * lightRadius)/((vx - light.x - SIZE * lightRadius)*(vx - light.x - SIZE * lightRadius)+(vy - light.y - SIZE * lightRadius)*(vy - light.y - SIZE * lightRadius)));

				// Light banding because cool
				if(bandedEnabled){
					if(scal>1.2){
						scal=1.2;
					}
					else if(scal>1){
						scal=1;
					}
					else if(scal>0.8){
						scal=0.8;
					}
					else if(scal>0.6){
						scal=0.6;
					}
					else {
						scal=lightMinimum;
					}
				}
				else {
					if(scal>1.2){
						scal=1.2;
					}
					else if(scal<lightMinimum){
						scal=lightMinimum;
					}
				}

				//if(scal !=lightMinimum){
					scal *=lightIntensity;
				//}

				c=GraphicUtils.scalehex(c, scal);

				var debris:Bool=(dx + dy>SIZE * debrisField * SIZE * debrisField);
				var cloud:Bool=(Math.random()<cloudDensity);

				if((debrisEnabled && debris)){

					g=overlay;

					c=GraphicUtils.scalehex(Math.random()>0.5?COL_NORMAL:COL_WET, scal *(0.75 + Math.random()* 0.25));
				}
				else if((cloudsEnabled && cloud)){
					g=overlay;

					c=GraphicUtils.scalehex(COL_SEA, scal *(1.25 + Math.random()* 0.25));
				}
				else {
					g=planet;
				}

				/*g.beginFill(c);
				g.drawRect(vertices[x][y].x			, vertices[x][y].y			, 1, 1);
				g.endFill();*/

				g.lineStyle();
				//g.lineStyle(1, 0x000000, 0.1);// uncomment for tile borders
				g.beginFill(c);
				g.moveTo((vertices[x][y].x			) * pixelScale, (vertices[x][y].y			) * pixelScale);
				g.lineTo((vertices[x][y + 1].x		) * pixelScale, (vertices[x][y + 1].y		) * pixelScale);
				g.lineTo((vertices[x + 1][y + 1].x	) * pixelScale, (vertices[x + 1][y + 1].y	) * pixelScale);
				g.lineTo((vertices[x + 1][y].x		) * pixelScale, (vertices[x + 1][y].y		) * pixelScale);
				g.lineTo((vertices[x][y].x			) * pixelScale, (vertices[x][y].y			) * pixelScale);
				g.endFill();
				//g.lineStyle();
			}
		}

	}

}